# Uml2adoc

A python script used to generate adoc-fragments by analysing comments in plantuml files.
Used by the OSC2.0 DomainModel group.

## Subfolder _'uml2adoc'_
This folder contains the script itself, an example config.yml and additional files needed.

## Subfolder _'examples'_
This folder contains some plantuml-files with added adoc-comments for test purposes.

## Subfolder _'guide'_
This folder contains a brief guide on how to document plantuml files.

## Subfolder _'antora_extension'_
This folder contains the Antora extension so the script can run within Antora directly.

## Configuration
Call the script with the path to your configuration yaml file:
`python uml2adoc.py "path/to/uml2adoc_config.yml"`

The *configuration yaml* shall contain the following entries:

* **`uml_inpath: "../examples/"`**

    Path where the UML files reside.

* **`adoc_outpath: "../_generated/"`**

    Path where the generated adoc-fragments will be stored.
    CAUTION: The content of this path will be deleted every time Uml2adoc is started!!!

* _OPTIONAL:_ **`linktype: "antora"`**

    Configure the type of link to be used:
    * antora: "xref:module:filename[linkname]"
    * default: "<<anchor, anchorname>>"

* _OPTIONAL:_ **`linkmodulename: "modulename"`**

    Specify a dedicated module name to be used for antora-style links


* **`logging: True`**

    Set to True to activate file and stdout logging, False for no logging.

*  **`logfile: "uml2adoc.log"`**

    Logging file name.

* _OPTIONAL:_ **`source_outpath: "../_generated/source/"`**

    Path where the generated source code files will be stored.
    CAUTION: The content of this path will be deleted every time Uml2adoc is started!!!


* _OPTIONAL:_ **`plantuml_jar: "/path/to/plantuml.jar"`**

    Path to the plantuml.jar for svg diagram creation.
    If set, the gui for plantuml is called at the beginning to create svg files.
This step is not necessary for adoc-fragment creation. Only for debugging.

* _OPTIONAL:_ **`png_outpath: "../_generated/png/"`**

    Path where the svg files created by planuml.jar will be stored.
    CAUTION: The content of this path will be deleted every time Uml2adoc is started!!!


An example configuration file can be found in the uml2adoc folder.

Note: Uml2adoc uses 'pyyaml' to read the configuration file. You may have to install pyyaml using `pip install pyyaml`

## Antora
### site.yml (generator)
To use the Antora extension, add it to the site.yml of the project and configure the workdir (i.e. the directory that is used within the docker container).

Note that the default workdir of the image is 'osc2'.

```yml
antora:
  extensions:
    - require: './uml2adoc/antora_extension/uml2adoc-integration.js'
      workdir: osc2 #now obsolete with new gitlab pipeline
```

NOTE: The workdir is optional only required if the pipeline or the container checks out the content into a subfolder.
If the Antora image runs directly in the root folder, omit the workdir attribute!

### antora.yml (project)
In addition to the general activation of the extension, it needs specific attributes for configuration.
Although these can be set in the generator, doing so would apply the same attributes to all components and versions.

This, typically, leads to errors and is not future proof. 
Instead, these setting should best be applied and updated in the respective antora.yml file(s).

```yml
asciidoc:
  attributes:
    uml2adoc: true
    uml2adoc_config: 'path/to/dir/in/generator/uml2adoc_config.yml'
    uml2adoc_legends_path: 'path/to/legend.plantuml' #this is the path within this version in relation to the antora.yml file. 
```
