'use strict'
var spawnSync = require("child_process").spawnSync;
var fs = require("fs");
const path = require("path")
const File = require('./file')

module.exports.register = function ({config}) {
    let {workdir} = config
    this
      .on('contentAggregated', ({contentAggregate}) => {
        // console.log(fs.readdirSync(process.cwd()+"/"+workdir))
        workdir = workdir ? workdir + "/" : ""
        const targetPath = './'+workdir+'uml2adoc/uml2adoc'
        contentAggregate.forEach(v => {
            console.log(v.version)
            if (v.asciidoc && v.asciidoc.attributes && v.asciidoc.attributes.uml2adoc) {
                const config_file = v.asciidoc.attributes.uml2adoc_config ? workdir + v.asciidoc.attributes.uml2adoc_config : workdir + 'uml2adoc_config.yml'
                const files = {
                    fileList: v.files
                }
                const defaultOrigin = v.files[0].src.origin
                const splitAbsPath = v.files[0].src.abspath ? v.files[0].src.abspath.split("/") : null
                const abspathPrefix = splitAbsPath ? splitAbsPath.slice(0,splitAbsPath.indexOf("modules")).join("/")+"/" : null
                const legendsBasePath = `./plantuml_build/${v.name}/${v.version}`
                const pathOffset = v.asciidoc.attributes.uml2adoc_legends_path ? "" : "standard/"
                v.asciidoc.attributes.uml2adoc_legends_path = v.asciidoc.attributes.uml2adoc_legends_path ? v.asciidoc.attributes.uml2adoc_legends_path : 'modules/domain-model/partials/definition/entity_definitions/legend.plantuml'
                const legendsFolder = `${legendsBasePath}/${pathOffset}${v.asciidoc.attributes.uml2adoc_legends_path}`
                const legendsDir = path.dirname(legendsFolder)
                const legendsFile = v.files.find(file => file.src.path.includes(v.asciidoc.attributes.uml2adoc_legends_path))
                if (legendsFile) {
                    if (fs.existsSync(legendsDir)) {
                        fs.rmSync(legendsDir, { recursive: true, force: true });
                    }
                    fs.mkdirSync(legendsDir, {recursive: true})
                    fs.writeFileSync(legendsFolder, legendsFile.contents)
                    v.asciidoc.attributes['kroki-plantuml-include-paths'] = legendsBasePath
                }

                try{
                    fs.writeFileSync(targetPath+"/virtual_files.json", JSON.stringify(files), err => {
                        if(err) {
                            console.log(err)
                            return
                        }
                    })
                    console.log("Temporary file written")
                    const python = spawnSync('python3', [targetPath+'/uml2adoc.py',config_file, targetPath+"/virtual_files.json"])
                    console.log(python.stdout.toString())
                    console.log(python.stderr.toString())
                    console.log("Python finished")
                    deleteTemporaryFile(targetPath+"/virtual_files.json")
                    console.log("Temporary input file deleted")
                    let pythonOutput = JSON.parse(fs.readFileSync(targetPath+"/output.json").toString())
                    let newFiles = []
                    pythonOutput.fileList.forEach(x => {
                        const contents = x._contents
                        // const stat = Object.assign(new fs.Stats(), { mode: entry.mode, mtime: undefined, size: contents.byteLength })
                        const existingFile = v.files.find(file => file.src.path === x.src.path)
                        if (existingFile) {
                            existingFile.contents = contents ? Buffer.from(contents) : existingFile.contents
                        }
                        else {
                            let file = new File({
                                path: x.src.path,
                                contents: Buffer.from(contents),
                                src: x.src
                            })
                            file.src.origin = defaultOrigin
                            if (abspathPrefix) {
                                file.src.abspath = abspathPrefix+file.src.path
                            }

                            newFiles.push(file)

                        }
                    })
                    deleteTemporaryFile(targetPath+"/output.json")
                    console.log("Temporary output file deleted")
                    v.files = v.files.concat(newFiles).sort((a,b) =>
                    {
                        return a.path - b.path
                    })
                } catch(e){
                    console.log(e)
                }
            }
        })
      })
  }

  function deleteTemporaryFile(path) {
    fs.unlink(path, function (err) {
        if (err) {
          console.error(err);
        } else {
          console.log("File removed:", path);
        }
      });
  }