# -*- coding: utf-8 -*-

VERSION = "Uml2adoc V3.2.1"

# known pantuml entities
KEYWORD_CLASS  = "class"
KEYWORD_ENUM  = "enum"
KEYWORD_INTERFACE = "interface"
KEYWORD_ABSTRACT = "abstract"
KEYWORD_ENTITY = "entity"
KEYWORD_NONE = "None"

VALID_KEYWORDS = ( KEYWORD_CLASS, KEYWORD_ENUM, KEYWORD_INTERFACE, KEYWORD_ABSTRACT, KEYWORD_ENTITY)

# strings for text
TEXT_CLASS_STRUCT = "Struct"
TEXT_ENUM = "Enum"
TEXT_INTERFACE_ABSTRACT = "Abstract"
TEXT_ENTITY = "Actor"
TEXT_CLASS_ACTION = "Action"

# keywords relevant for adoc
ADOC_KEYWORD_BASIC = "-basic"
ADOC_KEYWORD_PROPERTY = "-property"
ADOC_KEYWORD_STATE = "-state"
ADOC_KEYWORD_METHOD = "-method"
ADOC_KEYWORD_MODIFIER = "-modifier"
ADOC_KEYWORD_DEFAULT = "-default"
ADOC_KEYWORD_IGNORE = "-fileignore"
ADOC_KEYWORD_TARGETPAGE = "-antora_targetpage"
ADOC_KEYWORD_PACKAGENAME = "-packagename"
ADOC_KEYWORD_TAG = "-tag"
ADOC_KEYWORD_IGNORE_FOR_OSC = "-ignore_for_osc"
ADOC_KEYWORD_NONE = "None"

# plantuml relations
UML_RELATION_EXTENSION_1 = "<|--"
UML_RELATION_EXTENSION_2 = "<|.."
UML_RELATION_COMPOSITION_1 = "*--"
UML_RELATION_COMPOSITION_2 = "*.."
UML_RELATION_AGGREGATION_1 = "o--"
UML_RELATION_AGGREGATION_2 = "o.."
UML_RELATION_CONNECTION_1 = "--"
UML_RELATION_CONNECTION_2 = ".."
UML_RELATION_SMALL_CONNECTION_1 = "-"
UML_RELATION_SMALL_CONNECTION_2 = "."

#UML_RELATIONS = (  UML_RELATION_EXTENSION_1, UML_RELATION_EXTENSION_2
#                 , UML_RELATION_COMPOSITION_1, UML_RELATION_COMPOSITION_2
#                 , UML_RELATION_AGGREGATION_1, UML_RELATION_AGGREGATION_2
#                 , UML_RELATION_CONNECTION_1, UML_RELATION_CONNECTION_2)

# elements that need special treatment
KEYWORD_SPECIAL_TOGETHER = "together"
KEYWORD_SPECIAL_STATIC = "{static}"
KEYWORD_SPECIAL_ABSTRACT = "{abstract}"


# strings and fragments for parsing and replacement
STR_COMMENT_ADOC = "/'adoc"
STR_COMMENT_ADOC_REPLACEMENT = "©"
STR_COMMENT_IGNORE = "/'"
STR_COMMENT_IGNORE_REPLACEMENT = "®"
STR_COMMENT_END = "'/"
STR_COMMENT_END_REPLACEMENT = "¦"
STR_ILLEGAL_COMMENT = " 'adoc"
STR_APOSTROPHE_REPLACEMENT = "!"
STR_NEWLINE = "\n"
STR_NEWLINE_REPLACEMENT = "Þ"
STR_HASH = "#"
STR_QUOTATION_MARK = "\""
STR_STATE_PREFIX = "__"
STR_STATE_PREFIX_TEXT = "var "
STR_EMPTY_DETAILS = "%"
STR_ENV_TWEAK = "UWAFEA_"
STR_LST_TWEAK = "UWAFLP__"

STR_BODY_BEGIN = "{"
STR_BODY_END = "}"
STR_GROUP_BEGIN = "["
STR_GROUP_END = "]"
STR_APOSTROPHE = "'"
STR_BASIC_SEPARATOR = "~~"
STR_TABLE_SEPARATOR = "|"

PLANTUML_SPECIAL_CHARACTERS = f"""{STR_BODY_BEGIN}
                                  {STR_BODY_END}
                                  {STR_APOSTROPHE}
                                  {STR_COMMENT_ADOC_REPLACEMENT}
                                  {STR_COMMENT_IGNORE_REPLACEMENT}
                                  {STR_COMMENT_END_REPLACEMENT}""" # need to be separated

# basic information
STR_BASIC_INSTANTIABLE = "instantiable:"
STR_BASIC_PARENT = "parent:"
STR_BASIC_PARENTS = "parents:"
STR_BASIC_DESCRIPTION = "description:"
STR_BASIC_CONTROLLED_STATES = "controlled_states:"
STR_BASIC_ACTION_ENDING = "action_ending:"
STR_BASIC_ACTION_ACTORS= "action_actors:"
STR_BASIC_INCLUDE = "include:"
STR_BASIC_EXAMPLES = "examples:"
STR_BASIC_IMPLEMENTATION = "implementation:"

# anchor suffixes
STR_ANCHOR_SUFFIX_BASIC = "-basic"
STR_ANCHOR_SUFFIX_INFO = "-info"
STR_ANCHOR_SUFFIX_STATES = "-states"
STR_ANCHOR_SUFFIX_INHERITED = "-inherited-properties"

# files and filename components
OSC_FILE_NAME_STANDARD = "standard.osc"
OSC_FILE_NAME_TYPES = "types.osc"
OSC_FILE_NAME_DOMAIN = "domain.osc"
DIR_NAME_BARE = "osc_partials/"
OSC_PART_HEADER = "header.standard.osc"
OSC_PART_IMPORTS = "imports.standard.osc"
OSC_PART_TYPES = "physical-types.standard.osc"
OSC_PART_OTHER = "other-structs.standard.osc"
FILE_NAMES_BARE = [OSC_PART_HEADER, OSC_PART_IMPORTS, OSC_PART_TYPES, OSC_PART_OTHER,] # "other-actions.standard.osc",]
FILE_NAMEPART_STRUCT = "struct"
FILE_ENDING_ADOC = ".adoc"
FILE_ENDING_UML = ".uml"
FILE_ENDING_PLANTUML = ".plantuml"
FILE_ENDING_OSC2 = ".osc"

# type of link to generate
LINKTYPE_ANTORA = "antora"

# antora integration and virtual files
ANTORA_START_FOLDER = "modules"