# -*- coding: utf-8 -*-
import os
import json

from helpers import constants as C
from helpers import filehelper
from helpers import uml_parser as p
from helpers import sanitizer as s
from helpers import virtualfilehelper as v

dic_of_actions = {}
dml_oscs = {
    C.OSC_PART_HEADER : [] ,
    C.OSC_PART_IMPORTS : [] ,
    C.OSC_PART_TYPES : [],
    C.OSC_PART_OTHER : []

} # Strings for domain model library files

virtual_files = {
    "fileList": []
}

def dump_element(info : tuple, umlfile: str, virtual: bool) -> None:
    tup, key, name, file = info
    normals, basics, states, source, methods, modifiers = tup
    dic_of_children, dic_of_connections, dic_of_links, dic_of_parents, dic_of_users = p.get_set_of_dictionaries()

#    filehelper.logprint(f"\n{key} {name}")
#    filehelper.logprint("---basics---")
#    filehelper.logprint(basics)
#    filehelper.logprint(f"---normals ---")
#    filehelper.logprint(normals)
#    filehelper.logprint("---states---")
#    filehelper.logprint(states)
#    filehelper.logprint("---source---")
#    filehelper.logprint(source)
#    filehelper.logprint("---methods---")
#    filehelper.logprint(methods)
#    filehelper.logprint("---modifiers---")
#    filehelper.logprint(modifiers)
#    filehelper.logprint("---l---")
#    filehelper.logprint(dic_of_links)

    #BASICS
    intro_str = ""
    basics_str = ""
    examples_str = ""
    implem_str = ""

    if basics or name in dic_of_parents or name in dic_of_users or name in dic_of_connections:
        body = []
        instantiable = []
        parent = []
        description = []
        controlled_states = []
        action_ending = []
        action_actors = []
        include = []
        examples = []
        implem = []
        basic = ""

        for index, item in enumerate(basics):
            if item == C.ADOC_KEYWORD_BASIC:
                basic = ""
            elif item.lower() == C.STR_BASIC_INSTANTIABLE:
                basic = C.STR_BASIC_INSTANTIABLE
                #instantiable.append("\n|")
            elif item.lower() == C.STR_BASIC_PARENT or \
                 item.lower() == C.STR_BASIC_PARENTS:
                basic = C.STR_BASIC_PARENT
            elif item.lower() == C.STR_BASIC_DESCRIPTION:
                basic = C.STR_BASIC_DESCRIPTION
            elif item.lower() == C.STR_BASIC_CONTROLLED_STATES:
                basic = C.STR_BASIC_CONTROLLED_STATES
                if controlled_states:
                    controlled_states.append(C.STR_BASIC_SEPARATOR)
            elif item.lower() == C.STR_BASIC_ACTION_ENDING:
                basic = C.STR_BASIC_ACTION_ENDING
                if action_ending:
                    action_ending.append(C.STR_BASIC_SEPARATOR)
            elif item.lower() == C.STR_BASIC_ACTION_ACTORS:
                basic = C.STR_BASIC_ACTION_ACTORS
                if action_actors:
                    action_actors.append(C.STR_BASIC_SEPARATOR)
            elif item.lower() == C.STR_BASIC_INCLUDE:
                basic = C.STR_BASIC_INCLUDE
            elif item.lower() == C.STR_BASIC_EXAMPLES:
                basic = C.STR_BASIC_EXAMPLES
            elif item.lower() == C.STR_BASIC_IMPLEMENTATION:
                basic = C.STR_BASIC_IMPLEMENTATION
            elif basic == C.STR_BASIC_INSTANTIABLE:
                instantiable.append(item)
            elif basic == C.STR_BASIC_PARENT:
                parent.append(item)
            elif basic == C.STR_BASIC_DESCRIPTION:
                description.append(item)
            elif basic == C.STR_BASIC_CONTROLLED_STATES:
                controlled_states.append(item)
            elif basic == C.STR_BASIC_ACTION_ENDING:
                action_ending.append(item)
            elif basic == C.STR_BASIC_ACTION_ACTORS:
                action_actors.append(item)
            elif basic == C.STR_BASIC_INCLUDE:
                include.append(item)
            elif basic == C.STR_BASIC_EXAMPLES:
                examples.append(item)
            elif basic == C.STR_BASIC_IMPLEMENTATION:
                implem.append(item)

        intro_str = " ".join(description)

        if include:
            includestr = "".join(include)
            intro_str += f"\n\ninclude::{includestr}[]\n"

        if examples:
            examples_joined = "".join(examples)
            examples_str = f"\n\ninclude::{examples_joined}[]\n"

        if implem:
            implem_joined = "".join(implem)
            implem_str = f"\n\ninclude::{implem_joined}[]\n"

        for i, b in enumerate(body):
            if b in dic_of_links:
                body[i] = add_link(b)

        if instantiable:
            basics_str += "\n|*Instantiable* | "
            basics_str += ", ".join(instantiable)

        if not name in dic_of_parents:
            dic_of_parents[name] = parent

        basics_str += add_basic(name, dic_of_parents, "Parents")
        basics_str += add_basic(name, dic_of_children, "Children")
        basics_str += add_basic(name, dic_of_users, "Used by")
        basics_str += add_basic(name, dic_of_connections, "Has connection to")
        basics_str += add_action_basic(controlled_states, "Controlled states")
        basics_str += add_action_basic(action_ending, "Action ending")
        basics_str += add_action_basic(action_actors, "Actors")

        if basics_str:
            basics_str = "\n[cols=\"12,40\"]\n|===" + basics_str + "\n|==="

    #NORMAL AND PROPERTIES
    normals_str = ""

    # heading and table headers
    heading = "Members"
    header1 = "Member"
    header2 = "Comment"
    if key == C.KEYWORD_ENUM:
        heading = "Values"
        header1 = "Value"
    if C.ADOC_KEYWORD_PROPERTY in normals:
        heading = "Parameters"
        header1 = "Parameter"
        header2 = "Type | Mandatory | Description"

    if (normals):
        body = []
        body_str = ""
        if C.ADOC_KEYWORD_PROPERTY in normals:
            prop = False
            for item in normals:
                if item == C.ADOC_KEYWORD_PROPERTY:
                    prop = True
                    continue
                if prop and "\n|" in item:
                    prop = False
                    body.append(item)
                else:
                    if ";;" in item:
                        filehelper.log_error(f"found ';;' in {name} ('{item}')", umlfile)
                        item = item.replace(";;", ";")
                    subitems = item.split(";")
                    for i, subitem in enumerate(subitems):
                        subitems[i] = f"{add_link(subitem)}"
                    body.append("|".join(subitems))

            for i, b in enumerate(body):
                if b in dic_of_links:
                    body[i] = add_link(b)

            body_str = s.remove_optional_default(" ".join(body).replace(" \n", "\n"))

        else:

            for i, n in enumerate(normals):
                if n in dic_of_links:
                    normals[i] = add_link(n)

            body_str = " ".join(normals).replace(" \n", "\n")

        if len(normals) > 1:
            normals_filedump(name, key, normals)

            colwidth =  "12,40" if header2 == "Comment" else "12,10,5,40"
            normals_str = f"[cols=\"{colwidth}\"]\n[%header]\n|===\n| {header1} | {header2}"
            normals_str += s.untweak_p(body_str)
            normals_str +="\n|==="


    #STATES
    states_str = ""

    if (states):
        state = False
        body = []
        if ";" in states:
            filehelper.log_error(f"found ';;' in {name} state", umlfile)
            states.remove(";")
        for item in states:
            if item == C.ADOC_KEYWORD_STATE:
                state = True
                continue
            if state and ":" in item:
                state = False
                body.append("\n|")
                body.append(item.replace(":", "|"))
            else:
                subitems = item.split(";")
                for i, subitem in enumerate(subitems):
                    subitems[i] = f"{add_link(subitem)}"
                body.append("|".join(subitems))

        for i, b in enumerate(body):
            if b in dic_of_links:
                body[i] = add_link(b)

        states_str = "[cols=\"12,10,5,40\"]\n[%header]\n|===\n| Variable | Type | Mandatory | Description"
        states_str += " ".join(body).replace(" \n", "\n")
        states_str +="\n|==="

    #INHERITED PROPERTIES
    inherited_properties_str = get_parents_properties_str(name, dic_of_parents)

    if inherited_properties_str:
        inherited_properties_str = "[cols=\"12,40\"]\n[%header]\n|===\n| Parent | Inherited parameters and variables" + inherited_properties_str + "\n|===\n"

    methods_str = ""
    if methods:
        for methodname, methodpath in methods:
            methods_str += f"\n=== Method {methodname}\n\ninclude::{methodpath}[]\n"

    modifiers_str = ""
    if modifiers:
        for modifiername, modifierpath in modifiers:
            modifiers_str += f"\n=== Modifier {modifiername}\n\ninclude::{modifierpath}[]\n"

    isAction = basics and (action_ending or controlled_states)
    element_filedump(name, key, intro_str, basics_str, heading, normals_str, states_str, inherited_properties_str, examples_str, implem_str, methods_str, modifiers_str, isAction, file, umlfile, virtual)

    if source or modifiers:
        file = os.path.splitext(file)[0]
        parents = dic_of_parents[name] if name in dic_of_parents else ""
        source_filedump(name, key, source, parents, methods, modifiers, isAction, file, virtual)

    if basics and action_ending:
        dic_of_actions[name] = (" ".join(controlled_states), " ".join(action_ending))


def get_parents_properties_str(name: str, dic_of_parents: dict) -> str:
    return_str = ""

    parents_list = []
    temp_list = []

    links = p.get_dic_of_links()
    parentprops = p.get_dic_of_parentproperties()

    if name in dic_of_parents:
        for item in dic_of_parents[name]:
            temp_list.append(item)
            parents_list.append(item)

    while temp_list:
        temp_parent_list = []
        for tempitem in temp_list:
            if tempitem in dic_of_parents:
                temp_parent_list = dic_of_parents[tempitem]
        if temp_parent_list:
            parents_list.extend(temp_parent_list)
        temp_list = temp_parent_list

    if parents_list:
        parents_list = list(set(parents_list))
        parents_list.sort()
        for i, item in enumerate(parents_list):
            if item in links and item in parentprops:
                parents_list[i] = add_link(item)
                properties = []
                for prop in parentprops[item]:
                    if not prop[0] == C.STR_EMPTY_DETAILS:
                        linkitem = links[item]
                        property_link = linkitem.replace("sec-", "tab-")
                        property_link += C.STR_ANCHOR_SUFFIX_STATES if C.STR_STATE_PREFIX in prop[0] else C.STR_ANCHOR_SUFFIX_INFO
                        property_name = prop[0].replace(C.STR_STATE_PREFIX, "")
                        if filehelper.get_linktype() == C.LINKTYPE_ANTORA:
                            targetpages = p.get_dic_of_targetpages()
                            if item in targetpages:
                                targetpage = targetpages[item]
                                if len(targetpage) > 0:
                                    properties.append(f"xref:{targetpage}.adoc#{property_link}[{property_name}]")
                                else:
                                    properties.append(property_name)
                            else:
                                properties.append(property_name)
                        else:
                            properties.append(f"<<{property_link}, {property_name}>>")
                if properties:
                    return_str += f"\n| {parents_list[i]} | "
                    return_str += ", ".join(properties)
    return return_str


def add_basic(name : str, dic : dict, header : str) -> str:
    returnstr = ""
    temp_list = []

    if name in dic:
        for item in dic[name]:
            temp_list.append(item)

    if temp_list:
        temp_list = list(set(temp_list))
        temp_list.sort()
        for i, item in enumerate(temp_list):
            temp_list[i] = add_link(item)
        returnstr = f"\n|*{header}* | "
        returnstr += ", ".join(temp_list)
    return returnstr


def add_action_basic(textlist : list, header : str) -> str:
    returnstr = ""
    if textlist:
        for i, item in enumerate(textlist):
            textlist[i] = add_link(item)
        returnstr = f"\n|*{header}* | "
        returnstr += " ".join(textlist)
        returnstr = returnstr.replace(f" {C.STR_BASIC_SEPARATOR} ", "; +\n")
    return returnstr


def add_link(item: str) -> str:
    links = p.get_dic_of_links()
    if item in links:
        linkitem = links[item]

        if filehelper.get_linktype() == C.LINKTYPE_ANTORA:
            targetpages = p.get_dic_of_targetpages()

            if item in targetpages:
                # modulename = filehelper.get_linkmodule()
                # if len(modulename) > 0:
                targetpage = targetpages[item]
                if len(targetpage) > 0:
                    return f"xref:{targetpage}.adoc#{linkitem}[{item}]"
            return item
        else:
            return f"<<{linkitem}, {item}>>"
    else:
        return item


def get_topic_anchor_suffix(umlfile: str, key: str, name : str) -> str:
    return os.path.splitext(umlfile)[0].casefold() + f"-{key}-{name.casefold()}"


def list_filedump(virtual : bool, thelist : list, heading : str, outname : str, suffix : str =".adoc") -> None:
    """
    dump the list into a file. List is a tuple of name and path
    """
    if thelist:
        thelist = sorted(list(set(thelist)), key = lambda s: s[0].casefold())

        if virtual:
            extname = "adoc"
            file = filehelper.get_adoc_outpath()+outname+"."+extname
            file_parts = [outname,extname]
            v_file = create_virtual_file_base(file,file_parts,filehelper.get_adoc_outpath())
            body = f"// This file was created automatically by {C.VERSION} (cp_a).\n// DO NOT EDIT!\n\n"
            outname = s.untweak(outname.replace("_", "-"))
            body += "[#sec-{outname}]\n= {heading}\n\n"
            for n, f in thelist:
                    if f:
                        #body += "* xref:{f}#[{n}]\n"
                        body += "* {s.untweak(add_link(s.untweak(n)))}\n"
                    #else:
                        #body += "* {n}\n"
                        #body += "* {n}\n"

            v_file['_contents'] = body
            virtual_files['fileList'].append(v_file)


        else:
            with open(f"{filehelper.get_adoc_outpath()}{outname}.adoc", 'a', encoding="latin-1") as fobj:
                fobj.write(f"// This file was created automatically by {C.VERSION} (cp_b).\n// DO NOT EDIT!\n\n")
                #fobj.write(f":outfilesuffix: {suffix}\n\n")
                #fobj.write("ifndef::includedir[]\n:includedir: ./\nendif::includedir[]\n\n")
                outname = s.untweak(outname.replace("_", "-"))
                fobj.write(f"[#sec-{outname}]\n= {heading}\n\n")
                for n, f in thelist:
                    if f:
                        #fobj.write(f"* xref:{f}#[{n}]\n")
                        fobj.write(f"* {s.untweak(add_link(s.untweak(n)))}\n")
                    #else:
                        #fobj.write(f"* {n}\n")
                        #fobj.write(f"* {n}\n")


def element_filedump(name: str, key : str, intro : str, basics : str, heading: str, normals : str, states : str, inherited_properties : str, examples : str, implem : str, methods : str, modifiers : str, isAction : bool, file: str, umlfile: str, virtual: bool) -> None:

    text = get_entity_name_from_key(name, key, isAction)

    if virtual:
        file = s.untweak(file)
        file_parts = file.split(".")
        v_file = create_virtual_file_base(file,file_parts,filehelper.get_adoc_outpath())

        anchorstr = s.untweak(get_topic_anchor_suffix(umlfile, key, name))
        body = f"// This file was created automatically by {C.VERSION} (cp_c) using the plantuml file '{umlfile}'.\n// DO NOT EDIT!\n\n"
        body += f"[#sec-{anchorstr}]\n= {text} {s.untweak(name)}\n\n"
        # if filehelper.get_linktype() == C.LINKTYPE_ANTORA:
        #     body += "include::{bib}[]\n\n"
        if intro:
            body += intro
            body += "\n\n"
        if basics:
            body += "Basic information::\n+"
            body += f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_BASIC}]\n.Basic information of {text.lower()} {s.untweak(name)}\n"
            body += basics
        if normals:
            body += f"\n\n{heading}::\n+"
            body += f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_INFO}]\n.{text} {s.untweak(name)}\n"
            body += normals
        if states:
            body += "\n\nState variables::\n+"
            body += f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_STATES}]\n.State variables of {text.lower()} {s.untweak(name)}\n"
            body += states
        if inherited_properties:
            body += "\n\nInherited parameters and variables::\n+"
            body += f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_INHERITED}]\n.Inherited parameters and variables of {text.lower()} {s.untweak(name)}\n"
            body += inherited_properties
        if methods:
            body += "\n\n== Methods\n"
            body += methods
        if modifiers:
            body += "\n\n== Modifiers\n"
            body += modifiers
        if examples:
            body += "\n\n== Examples\n"
            body += examples
        if implem:
            body += "\n\n== Implementation details\n"
            body += implem

        body += "\n"
        v_file['_contents'] = body
        virtual_files['fileList'].append(v_file)

    else:
        with open(f"{filehelper.get_adoc_outpath()}{s.untweak(file)}", "a", encoding="utf-8") as fobj:
            anchorstr = s.untweak(get_topic_anchor_suffix(umlfile, key, name))
            fobj.write(f"// This file was created automatically by {C.VERSION} (cp_d) using the plantuml file '{umlfile}'.\n// DO NOT EDIT!\n\n")
            #fobj.write(":outfilesuffix: .adoc\n")
            #fobj.write("ifndef::includedir[]\n:includedir: ./\nendif::includedir[]\n\n")
            fobj.write(f"[#sec-{anchorstr}]\n= {text} {s.untweak(name)}\n\n")

            # if filehelper.get_linktype() == C.LINKTYPE_ANTORA:
            #     fobj.write("include::{bib}[]\n\n")
            if intro:
                fobj.write(intro)
                fobj.write("\n\n")
            if basics:
                fobj.write("Basic information::\n+")
                fobj.write(f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_BASIC}]\n.Basic information of {text.lower()} {s.untweak(name)}\n")
                fobj.write(basics)
            if normals:
                fobj.write(f"\n\n{heading}::\n+")
                fobj.write(f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_INFO}]\n.{text} {s.untweak(name)}\n")
                fobj.write(normals)
            if states:
                fobj.write("\n\nState variables::\n+")
                fobj.write(f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_STATES}]\n.State variables of {text.lower()} {s.untweak(name)}\n")
                fobj.write(states)
            if inherited_properties:
                fobj.write("\n\nInherited parameters and variables::\n+")
                fobj.write(f"\n[#tab-{anchorstr}{C.STR_ANCHOR_SUFFIX_INHERITED}]\n.Inherited parameters and variables of {text.lower()} {s.untweak(name)}\n")
                fobj.write(inherited_properties)
            if methods:
                fobj.write("\n\n== Methods\n")
                fobj.write(methods)
            if modifiers:
                fobj.write("\n\n== Modifiers\n")
                fobj.write(modifiers)
            if examples:
                fobj.write("\n\n== Examples\n")
                fobj.write(examples)
            if implem:
                fobj.write("\n\n== Implementation details\n")
                fobj.write(implem)

            fobj.write("\n")

list_of_sources = []


def source_filedump(name : str, key: str, source : list, parents: list, methods: list, modifiers: list, isAction : bool, file : str, virtual: bool) -> None:

    global dml_oscs

    if filehelper.get_source_outpath() == ""  or name in p.get_dic_of_oscignore() or key == C.KEYWORD_ENUM:
        return

    outfilename_adoc = f"{filehelper.get_source_outpath()}{file}_{C.FILE_NAMEPART_STRUCT}{C.FILE_ENDING_ADOC}"

    list_of_sources.append((name, outfilename_adoc))

    outstrings_adoc = []
    added = False

    outstrings_adoc.append(f"\nstruct {s.untweak(name)}")

    tdic = p.get_dic_of_tagnames()

    if name in tdic:
        dml_oscs[C.OSC_PART_OTHER].append(f"\n# tag::{tdic[name]}[]")

    dml_oscs[C.OSC_PART_OTHER].append(f"\n{get_entity_name_from_key(name, key, isAction).casefold()} {get_package_name(name)}{s.untweak(name)}")

    if parents:
        parents = set(parents)
        if "-" in parents or "none" in parents or "no" in parents:
            pass
        else:
            outstrings_adoc.append(" extends ")

            parentnames = ",".join(parents)
            parentpack = "" if "." in parentnames else get_package_name(parentnames)
            dml_oscs[C.OSC_PART_OTHER].append(f" inherits {parentpack}")

            outstrings_adoc.append(parentnames)
            dml_oscs[C.OSC_PART_OTHER].append(parentnames)
    outstrings_adoc.append(":\n")
    dml_oscs[C.OSC_PART_OTHER].append(":\n")

    max_length_0 = 0
    max_length_1 = 0
    for item in source:
        if len(item[0]) > max_length_0:
            max_length_0 = len(item[0])
        extra_space = 6 if (len(item) > 4 and len(item[4]) > 0) else 0
        acutal_length_1 = len(repl_link(item[1])) + extra_space
        if acutal_length_1 > max_length_1:
            max_length_1 = acutal_length_1

    for item in source:
        if item[0][0] == C.STR_EMPTY_DETAILS:
            dml_oscs[C.OSC_PART_OTHER].pop()
            dml_oscs[C.OSC_PART_OTHER].append("\n")
        elif not item[0][0] == '_':
            item_1 = repl_link(s.untweak_p(item[1]))
            outstrings_adoc.append(f"    {item[0]}: {item_1}\n")
            dml_oscs[C.OSC_PART_OTHER].append(f"    {item[0]}: ")
            space = max_length_0 - len(item[0])
            for k in range(0, space):
                dml_oscs[C.OSC_PART_OTHER].append(" ")
            dml_oscs[C.OSC_PART_OTHER].append(f"{item_1}")
            space = max_length_1
            if len(item[4]) > 0:
                dml_oscs[C.OSC_PART_OTHER].append(" with:")
                space -= 6
            space -= len(repl_link(item_1))
            for k in range(0, space):
                dml_oscs[C.OSC_PART_OTHER].append(" ")
            dml_oscs[C.OSC_PART_OTHER].append(f"  # {repl_mand(item[3])} {repl_link(item[2], True).strip()}\n")
            if len(item[4]) > 0:
                dml_oscs[C.OSC_PART_OTHER].append(f"        keep(default it == {item[4]})\n")
            added = True
    if added:
        outstrings_adoc.append("\n")
        dml_oscs[C.OSC_PART_OTHER].append("\n")
        added = False
    for item in source:
        if item[0][0] == '_':
            outstrings_adoc.append(f"    {item[0]}: {item[1]}\n")
            dml_oscs[C.OSC_PART_OTHER].append(f"    {item[0].replace(C.STR_STATE_PREFIX, C.STR_STATE_PREFIX_TEXT)}: {repl_link(item[1])} # {repl_link(item[2], True).strip()}\n")
            added = True
    if added:
        outstrings_adoc.append("\n")
        dml_oscs[C.OSC_PART_OTHER].append("\n")

    if methods:
        for method in methods:
            try:
                lines = []
                method_file = f"{filehelper.get_uml_inpath()}{method[1][3:]}"

                if virtual:
                    for v_file in v.get_virtual_files_with_path(method_file)['fileList']:
                        lines = [x.replace("\r","") + "\n" for x in bytes(v_file['_contents']['data']).decode('utf-8').split("\n")]
                else:
                    with open(method_file, 'r', encoding="utf-8") as fobj:
                        lines = fobj.readlines()

                for line in lines:
                    if line.find("def ") > -1:
                        dml_oscs[C.OSC_PART_OTHER].append(line.rstrip("\n") + " is undefined\n")
                        break
                    elif line.startswith("`"):
                        dml_oscs[C.OSC_PART_OTHER].append("    def " + line.strip("`\n")[line.find("."):] + " is undefined #TODO enhance method description\n")
                        break
                    else:
                        continue
            except:
                dml_oscs[C.OSC_PART_OTHER].append(f"\n# ERROR: '{method[1]}' missing\n\n")
                filehelper.log_error(f"Could not find file for method '{method[1]}'")

    if name in tdic:
        dml_oscs[C.OSC_PART_OTHER].pop()
        dml_oscs[C.OSC_PART_OTHER].append(f"# end::{tdic[name]}[]\n")

    for modifier, modifier_path in modifiers:
        try:
            lines = []
            modifier_file = f"{filehelper.get_uml_inpath()}{modifier_path[3:]}"

            if virtual:
                for v_file in v.get_virtual_files_with_path(modifier_file)['fileList']:
                    lines = [x.replace("\r","") + "\n" for x in bytes(v_file['_contents']['data']).decode('utf-8').split("\n")]
            else:
                with open(modifier_file, 'r', encoding="utf-8") as fobj:
                    lines = fobj.readlines()

            lines = s.extract_modifier(lines)
            for line in lines:
                dml_oscs[C.OSC_PART_OTHER].append(line)
        except:
            dml_oscs[C.OSC_PART_OTHER].append(f"\n\n# ERROR: '{modifier_path}' missing\n\n")
            filehelper.log_error(f"Could not find file for modifier '{modifier_path}'")



def normals_filedump(name : str, key: str, normals : list) -> None:

    global dml_oscs
    if filehelper.get_source_outpath() == "" or name in p.get_dic_of_oscignore():
        return

    if key == C.KEYWORD_ENUM:
        dml_oscs[C.OSC_PART_OTHER].append(f"\n{C.KEYWORD_ENUM} {get_package_name(name)}{name}: [")

        name_max_length = 0
        for i, n in enumerate(normals):
            if i > 0 and normals[i - 1] == "\n|" and len(repl_link(n)) > name_max_length:
                name_max_length = len(n)

        for i, n in enumerate(normals):
            if n == "\n|":
                dml_oscs[C.OSC_PART_OTHER].append("\n    ")
            elif n == "|":
                dml_oscs[C.OSC_PART_OTHER].append(", ")
                for k in range(0, name_max_length - len(repl_link(normals[i-1]))):
                    dml_oscs[C.OSC_PART_OTHER].append(" ")
                dml_oscs[C.OSC_PART_OTHER].append(" # ")
            elif (i < (len(normals) - 1)) and normals[i + 1] == "|":
                dml_oscs[C.OSC_PART_OTHER].append(f"{repl_link(n)}")
            else:
                dml_oscs[C.OSC_PART_OTHER].append(f"{repl_link(n)} ")

        dml_oscs[C.OSC_PART_OTHER].append("\n")

        index_last_occurence = len(dml_oscs[C.OSC_PART_OTHER]) - dml_oscs[C.OSC_PART_OTHER][::-1].index(", ") - 1
        if index_last_occurence > -1:
            dml_oscs[C.OSC_PART_OTHER][index_last_occurence] = "] "


def repl_link(text: str, verbose: bool = False) -> str:
    ret = text
    if "xref:" in text:
        try:
            if verbose:
                ret = "See definition of '" + text[text.find('[') + 1 : text.find(']')] + "'."
            else:
                ret = text[:text.find('xref:')] + text[text.find('[') + 1 : text.find(']')]
        except:
            filehelper.log_error("Could not replace link")
    elif "<<" and ">>" in text:
        ret = text[text.find(', ') + 2 : text.find('>>')]
    return ret


def repl_mand(text: str) -> str:
    ret = ""
    if text.casefold() == "no":
        ret = "Optional: "
    elif text.casefold() == "yes":
        ret = "Mandatory:"
    return ret


def paste_bare(virtual: bool) -> None:

    global dml_oscs
    bare_dir = f"{filehelper.get_uml_inpath()}{C.DIR_NAME_BARE}"
    # bares = filehelper.get_osc_partial_names(bare_dir) # Only works with local files, e.g. for running inside the content repository or for local development. DO NOT USE THIS WITH ANTORA (will not work as expected)
    bares = C.FILE_NAMES_BARE

    for bare_name in bares:
        bare_full_name = f"{bare_dir}{bare_name}"

        try:
            bare = []
            if virtual:
                for v_file in v.get_virtual_files_with_path(bare_full_name)['fileList']:
                    bare = [x.replace("\r","") + "\n" for x in bytes(v_file['_contents']['data']).decode('utf-8').split("\n")]
            else:

                if not os.path.exists(bare_full_name):
                    filehelper.log_error(f"Bare '{bare_name}' not found")
                    continue

                with open(bare_full_name, "r", encoding="utf-8") as fbare:
                    bare = fbare.readlines()

            bare.append("\n")
            for b in bare:
                if b.startswith("# tag::") or b.startswith("# end::"):
                    bare.remove(b)
            dml_oscs[bare_name].extend(bare)
        except:
            dml_oscs[bare_name].append(f"\n\n# ERROR: '{bare_name}' missing\n\n")
            filehelper.log_error(f"Could not find file for modifier '{bare_name}'")


def get_entity_name_from_key(name : str, key: str, isAction: bool) -> str:

    text = "?"
    stereotypes = p.get_dic_of_stereotypes()

    if name in stereotypes:
        text = stereotypes[name]
    else:
        if key == C.KEYWORD_ENUM:
            text = C.TEXT_ENUM
        elif key == C.KEYWORD_CLASS:
            if isAction:
                text = C.TEXT_CLASS_ACTION
            else:
                text = C.TEXT_CLASS_STRUCT
        elif key == C.KEYWORD_ENTITY:
            text = C.TEXT_ENTITY
        elif key == C.KEYWORD_INTERFACE or key == C.KEYWORD_ABSTRACT:
            text = C.TEXT_INTERFACE_ABSTRACT
    return text


def get_package_name(name: str) -> str:
    pdic = p.get_dic_of_packagenames()
    if name in pdic:
        return pdic[name] + "."
    else:
        return ""


def actions_filedump(virtual:bool) -> None:
    if dic_of_actions:

        if virtual:
            basename = "actions"
            extname = "adoc"
            file = filehelper.get_adoc_outpath()+basename+"."+extname
            file_parts = [basename,extname]
            v_file = create_virtual_file_base(file,file_parts,filehelper.get_adoc_outpath())
            body = f"// This file was created automatically by {C.VERSION} (cp_k).\n// DO NOT EDIT!\n\n"
            body += "[#tab-dm-all-actions]\n.Actions, their controlled states and end conditions\n"
            body += "[cols=\"12,30,30\"]\n[%header]\n|===\n|Action |Controlled states |Action ending\n"

            for a in sorted(dic_of_actions):
                body += f"|{s.untweak(add_link(s.untweak(a)))} |{dic_of_actions[a][0]} |{dic_of_actions[a][1]}\n"

            body += "|===\n"

            v_file['_contents'] = body
            virtual_files['fileList'].append(v_file)

        else:
            with open(f"{filehelper.get_adoc_outpath()}actions.adoc", "a", encoding="utf-8") as fobj:
                fobj.write(f"// This file was created automatically by {C.VERSION} (cp_l).\n// DO NOT EDIT!\n\n")
                fobj.write("[#tab-dm-all-actions]\n.Actions, their controlled states and end conditions\n")
                fobj.write("[cols=\"12,30,30\"]\n[%header]\n|===\n|Action |Controlled states |Action ending\n")

                for a in sorted(dic_of_actions):
                    fobj.write(f"|{s.untweak(add_link(s.untweak(a)))} |{dic_of_actions[a][0]} |{dic_of_actions[a][1]}\n")

                fobj.write("|===\n")


def dump_standard_osc(virtual: bool) -> None:
    dml_files = [C.OSC_FILE_NAME_STANDARD, C.OSC_FILE_NAME_TYPES, C.OSC_FILE_NAME_DOMAIN]

    filehelper.logprint("Writing domain model library files...")

    for outfilename in dml_files:
        try:

            filehelper.logprint(outfilename)

            if not virtual:
                for s in dml_oscs[C.OSC_PART_OTHER]:
                    s = s.replace("\r","")

            body = "".join(dml_oscs[C.OSC_PART_HEADER])

            if outfilename == C.OSC_FILE_NAME_STANDARD:
                body = body + "".join(dml_oscs[C.OSC_PART_IMPORTS])
            elif outfilename == C.OSC_FILE_NAME_TYPES:
                body = body + "".join(dml_oscs[C.OSC_PART_TYPES])
            elif outfilename == C.OSC_FILE_NAME_DOMAIN:
                body = body + "".join(dml_oscs[C.OSC_PART_OTHER])
            else:
                pass

            if virtual:
                v_file = create_virtual_file_base(outfilename, outfilename.split("."), path = filehelper.get_source_outpath())
                v_file['_contents'] = body.replace("\n","\r\n")
                virtual_files['fileList'].append(v_file)
            else:
                outfilename = filehelper.get_source_outpath()+outfilename
                with open(outfilename, "a", encoding="utf-8") as fs:
                    fs.write(body)
            filehelper.logprint("OK")
        except:
            filehelper.log_error("Could not write domain model library")


def dump_lists(virtual: bool) -> None:
    classes, enums, entities, interfaces, abstracts = p.get_set_of_lists()
    list_filedump(virtual, classes, "List of classes", "list_of_classes")
    list_filedump(virtual, enums, "List of enums", "list_of_enums")
    list_filedump(virtual, entities, "List of entities", "list_of_entities")
    interfaces = list(set(interfaces))
    list_filedump(virtual, interfaces, "List of interfaces", "list_of_interfaces")
    list_filedump(virtual, abstracts, "List of abstracts", "list_of_abstracts")
    list_filedump(virtual, list_of_sources, "List of source files", "list_of_sourcefiles")
    actions_filedump(virtual)

    #filehelper.logprint("\nStereotypes:")
    #stereotypes = p.get_dic_of_stereotypes()
    #filehelper.logprint(stereotypes)


def virtual_dump(path:str) -> None:
    with open(path+"output.json", "w") as jfile:
        json.dump(virtual_files,jfile)


def create_virtual_file_base(file:str, file_parts:list, path:str) -> dict:
    v_file = {
            'src': {
                "path": path+file,
                "basename": file,
                "stem": file_parts[0],
                "extname": "."+file_parts[1],
                "origin": {
                    "type": "uml2adoc"
                }
            }
        }
    return v_file