# -*- coding: utf-8 -*-
import os
import sys
import shutil
from os.path import join

import yaml

from helpers import constants as C

dic_config = {}

errorcount = 0

def load_config(configfile : str) -> bool:
    if os.path.exists(configfile):
        with open(configfile, 'r', encoding="utf-8") as fobj:
            global dic_config
            dic_config = yaml.safe_load(fobj).copy()
            return True
    else:
        print("Config file not found!")
        return False

def get_dic_config() -> dict:
    return dic_config

def set_dic_config(new_config) -> None:
    dic_config = new_config

def get_uml_inpath() -> str:
    return dic_config["uml_inpath"] if "uml_inpath" in dic_config else ""


def get_adoc_outpath() -> str:
    return dic_config["adoc_outpath"] if "adoc_outpath" in dic_config else ""


def get_source_outpath() -> str:
    return dic_config["source_outpath"] if "source_outpath" in dic_config else ""


def get_plantuml_jar() -> str:
    return dic_config["plantuml_jar"] if "plantuml_jar" in dic_config else ""


def get_png_outpath() -> str:
    return dic_config["png_outpath"] if "png_outpath" in dic_config else ""


def get_linktype() -> str:
    return dic_config["linktype"].casefold() if "linktype" in dic_config else ""


def get_linkmodule() -> str:
    return dic_config["linkmodulename"].casefold() if "linkmodulename" in dic_config else ""


def prepare_filestructure() -> None:
    if os.path.exists(dic_config["logfile"]):
        os.remove(dic_config["logfile"])

    logprint(dic_config)

    if get_adoc_outpath():
        clean_filestructure(get_adoc_outpath())
    if get_source_outpath():
        clean_filestructure(get_source_outpath())
    if get_png_outpath():
        clean_filestructure(get_png_outpath())


def clean_filestructure(path : str) -> None:
    if os.path.exists(path):
        for filename in os.listdir(path):
            fullfilename = os.path.join(path, filename)
            if os.path.isfile(fullfilename):
                os.remove(fullfilename)
    else:
        os.makedirs(path)


def logprint(message : str, end=" ") -> None:
    try:
        if dic_config["logging"]:
            print(message, end)

            with open(dic_config["logfile"], 'a') as fobj:
                print(message, end, file=fobj)
    except:
        # user configured no logging mode
        pass


def log_error( message : str, file : str = "", line : int = -1,) -> None:
    if line > -1:
        logprint(f"Error: {message} in line {line} of {file}!")
    elif file:
        logprint(f"Error: {message} in {file}")
    else:
        logprint(f"Error: {message}")
    global errorcount
    errorcount += 1


def log_summary() -> None :
    if (errorcount):
        logprint(f"{C.VERSION} finished with {errorcount} WARNINGS")
    else:
        logprint(f"{C.VERSION} finished OK")


def get_osc_partial_names(dir: str) -> list:

    ret_names = C.FILE_NAMES_BARE

    try:
        for root, dirs, files in os.walk(dir):
            for file in files:
                bare_file_name = os.path.splitext(file)[0] + os.path.splitext(file)[1]
                if not bare_file_name in ret_names:
                    ret_names.append(bare_file_name)
    except:
        pass

    return ret_names
