# -*- coding: utf-8 -*-

from helpers import constants as C
from helpers import filehelper as f


def sanitize(lines: list, filename : str) -> list:
    cleanlist = sanitize_lines(lines, filename)
    cleanlist = sanitize_comments(cleanlist)
    cleanlist = [s for s in cleanlist if s != "\n"]
    return cleanlist


def sanitize_lines(lines : list, filename : str) -> list:
    words = []

    for i, line in enumerate(lines):

        if C.STR_ILLEGAL_COMMENT in line:
            f.log_error(f"Found illegal keyword \"{C.STR_ILLEGAL_COMMENT}\"", filename, (i + 1))

        line = line.lstrip()

        if line.startswith("!"):
            f.logprint("Ignoring !-line")
            continue #ignore defines and includes

        for e in ["air", "wind", "fog", "clouds", "rain", "snow", "precipitation"]:
            line = line.replace(f"nment <|-- {e}", f"nment <|-- {C.STR_ENV_TWEAK}{e}")
            line = line.replace(f"ass {e} <<ACT", f"ass {C.STR_ENV_TWEAK}{e} <<ACT")

        line = line.replace("erty list of ", f"erty {C.STR_LST_TWEAK}")

        line = line.replace(C.STR_COMMENT_ADOC, C.STR_COMMENT_ADOC_REPLACEMENT)
        line = line.replace(C.STR_COMMENT_IGNORE, C.STR_COMMENT_IGNORE_REPLACEMENT)
        line = line.replace(C.STR_COMMENT_END, C.STR_COMMENT_END_REPLACEMENT)
        line = line.replace(C.KEYWORD_SPECIAL_STATIC, "static")
        line = line.replace(C.KEYWORD_SPECIAL_ABSTRACT, "abstract")

        linewords = "".join(char if char not in C.PLANTUML_SPECIAL_CHARACTERS else " " + char + " " for char in line).split()
        linewords.append(C.STR_NEWLINE_REPLACEMENT)
        linewords.append("\n")

        words.extend(linewords)

    in_comment = False
    for word in list(words):
        if word[0] == C.STR_HASH:
            words.remove(word)
        elif word == C.STR_COMMENT_ADOC_REPLACEMENT:
            in_comment = True
        elif word == C.STR_COMMENT_END_REPLACEMENT:
            in_comment = False
        elif len(word) > 4 and word.startswith(("--", "==")) and word.endswith(("--", "==")):
            words.remove(word)
        elif word.startswith("\"") and word.endswith("\"") and not in_comment:
            words.remove(word)

    return words



def untweak(n: str) -> str:
    if C.STR_ENV_TWEAK in n:
        return n.replace(C.STR_ENV_TWEAK, "")
    else:
        return n.replace(C.STR_ENV_TWEAK.casefold(), "")



def untweak_p(n: str) -> str:
    return n.replace(C.STR_LST_TWEAK, "list of ")


def sanitize_comments(wordlist: list) -> list:
    cleanlist = []

    ignoresingle = False # 'comment we do not want
    ignoremulti = False  # /'comment'/ we do not want
    adoc = False    # /'adoc... comment'/ we DO want

    for item in wordlist:
        if ignoresingle and item == "\n":
            ignoresingle = False
        if (ignoremulti or adoc) and item == C.STR_COMMENT_END_REPLACEMENT:
            ignoremulti = False
            adoc = False
        if item == C.STR_APOSTROPHE:
            if adoc:
                f.log_error("Ambiguous apostrophe")
            else:
                ignoresingle = True
        if item == "note":
            ignoresingle = True
        if item == C.STR_COMMENT_IGNORE_REPLACEMENT:
            ignoremulti = True
        if item == C.STR_COMMENT_ADOC_REPLACEMENT:
            adoc = True

        if not ignoresingle and not ignoremulti:
            cleanlist.append(item)

    return cleanlist


def remove_optional_default(text: str) -> str:
    ret_str = ''
    pipe_count = 0

    for c in text:
        if c == '\n':
            pipe_count = 0
        elif c == '|':
            pipe_count += 1
        if pipe_count < 5:
            ret_str += c
    return ret_str


def extract_modifier(lines : list) -> list:

    modifier_lines = []
    lookout = False
    dump = False
    for line in lines:
        if lookout and line.startswith("----"):
            if dump:
                dump = False
                lookout = False
            else:
                dump = True
                modifier_lines.append("\n")
            continue
        if line.startswith("ifdef::modifier_definition[]"):
            lookout = True
            continue
        if dump:
            modifier_lines.append(line)

    return modifier_lines
