# -*- coding: utf-8 -*-
import os

from helpers import constants as C
from helpers import filehelper as f
from helpers import sanitizer as s
from helpers import dumper as d

list_of_classes = []
list_of_enums = []
list_of_interfaces = []
list_of_abstracts = []
list_of_entities = []

dic_of_links = {}
dic_of_targetpages = {}
dic_of_packagenames = {}
dic_of_tagnames = {}
dic_of_oscignore = {}
dic_of_parents = {}
dic_of_children = {}
dic_of_users = {}
dic_of_connections = {}
dic_of_parentproperties = {}

dic_of_stereotypes = {}


def parse_uml_diagram_for_links(diagram: list, filename: str, file: str) -> None:

    f.logprint(filename)

    for word in diagram:
        if word.startswith("<<") and word.endswith(">>"):
            diagram.remove(word)

    parse_for_links(diagram, filename)
    parse_for_relationships(diagram, filename)

    detected_name = ""
    collected_items = []

    collect = C.KEYWORD_NONE
    first = False
    together = False
    together_paused = False

    #targetpages
    for index, item in enumerate(diagram):

        if collect in C.VALID_KEYWORDS:
            if first:
                first = False
                # skip name
                continue
            if item == C.STR_BODY_BEGIN:
                if together:
                    together_paused = together
                    together = False
                continue
            elif item == C.STR_BODY_END:
                if together:
                    together = False
                    continue

                if f.get_linktype() == C.LINKTYPE_ANTORA:
                    parse_items_for_targetpages(collected_items, detected_name)

                detected_name = ""
                collected_items.clear()
                collect = C.KEYWORD_NONE

                if together_paused:
                    together = together_paused
                    together_paused = False
            else:
                collected_items.append(item)
            continue

        if item.casefold() == C.KEYWORD_CLASS:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_CLASS
                detected_name = diagram[index + 1]
                first = True
            continue

        if item.casefold() == C.KEYWORD_ENUM:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_ENUM
                detected_name = diagram[index + 1]
                first = True
            continue

        if item.casefold() == C.KEYWORD_ENTITY:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_ENTITY
                detected_name = diagram[index + 1]
                first = True
            continue

        if item.casefold() == C.KEYWORD_INTERFACE:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_INTERFACE
                detected_name = diagram[index + 1]
                first = True
            continue

        if item.casefold() == C.KEYWORD_ABSTRACT:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_ABSTRACT
                detected_name = diagram[index + 1]
                first = True
            continue

        if item.casefold() == C.KEYWORD_SPECIAL_TOGETHER:
            if diagram[index + 1] == C.STR_BODY_BEGIN:
                together = True
                together_paused = False
            continue


def parse_uml_diagram(diagram: list, filename: str, file: str, virtual=False) -> None:

    f.logprint(file)
#    for li in dic_of_links:
#        f.logprint(f"{li} : {dic_of_links[li]}")

    #f.logprint("Parents:")
    #f.logprint(dic_of_parents)
    #f.logprint("Children:")
    #f.logprint(dic_of_children)
    #f.logprint("Users:")
    #f.logprint(dic_of_users)
    #f.logprint("Connections:")
    #f.logprint(dic_of_connections)

    detected_name = ""
    collected_items = []

    collect = C.KEYWORD_NONE
    first = False
    together = False
    together_paused = False

    for index, item in enumerate(diagram):

        if collect in C.VALID_KEYWORDS:
            if first:
                first = False
                # skip name
                continue
            if item == C.STR_BODY_BEGIN:
                if together:
                    together_paused = together
                    together = False
                continue
            elif item == C.STR_BODY_END:
                if together:
                    together = False
                    continue

                tup = parse_items(collected_items, detected_name, file)

                info4print = (tup, collect, detected_name, outfilename)
                d.dump_element(info4print, file, virtual)

                detected_name = ""
                collected_items.clear()
                collect = C.KEYWORD_NONE

                if together_paused:
                    together = together_paused
                    together_paused = False
            else:
                collected_items.append(item)
            continue

        if item.casefold() == C.KEYWORD_CLASS:
            if diagram[index + 2] == C.STR_BODY_BEGIN or diagram[index + 3] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_CLASS
                detected_name = diagram[index + 1]
                outfilename = f"{filename.casefold()}_{collect}_{detected_name.casefold()}{C.FILE_ENDING_ADOC}"
                list_of_classes.append((detected_name, outfilename))
                if diagram[index + 3] == C.STR_BODY_BEGIN:
                    register_stereotype(detected_name, diagram[index + 2])
                first = True
            continue

        if item.casefold() == C.KEYWORD_ENUM:
            if diagram[index + 2] == C.STR_BODY_BEGIN or diagram[index + 3] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_ENUM
                detected_name = diagram[index + 1]
                outfilename = f"{filename.casefold()}_{collect}_{detected_name.casefold()}{C.FILE_ENDING_ADOC}"
                list_of_enums.append((detected_name, outfilename))
                if diagram[index + 3] == C.STR_BODY_BEGIN:
                    register_stereotype(detected_name, diagram[index + 2])
                first = True
            continue

        if item.casefold() == C.KEYWORD_ENTITY:
            if diagram[index + 2] == C.STR_BODY_BEGIN or diagram[index + 3] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_ENTITY
                detected_name = diagram[index + 1]
                outfilename = f"{filename.casefold()}_{collect}_{detected_name.casefold()}{C.FILE_ENDING_ADOC}"
                list_of_entities.append((detected_name, outfilename))
                if diagram[index + 3] == C.STR_BODY_BEGIN:
                    register_stereotype(detected_name, diagram[index + 2])
                first = True
            continue

        if item.casefold() == C.KEYWORD_INTERFACE:
            if diagram[index + 2] == C.STR_BODY_BEGIN or diagram[index + 3] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_INTERFACE
                detected_name = diagram[index + 1]
                outfilename = f"{filename.casefold()}_{collect}_{detected_name.casefold()}{C.FILE_ENDING_ADOC}"
                list_of_interfaces.append((detected_name, outfilename))
                if diagram[index + 3] == C.STR_BODY_BEGIN:
                    register_stereotype(detected_name, diagram[index + 2])
                first = True
            continue

        if item.casefold() == C.KEYWORD_ABSTRACT:
            if diagram[index + 2] == C.STR_BODY_BEGIN or diagram[index + 3] == C.STR_BODY_BEGIN:
                collect = C.KEYWORD_ABSTRACT
                detected_name = diagram[index + 1]
                outfilename = f"{filename.casefold()}_{collect}_{detected_name.casefold()}{C.FILE_ENDING_ADOC}"
                list_of_abstracts.append((detected_name, outfilename))
                if diagram[index + 3] == C.STR_BODY_BEGIN:
                    register_stereotype(detected_name, diagram[index + 2])
                first = True
            continue

        if item.casefold() == C.KEYWORD_SPECIAL_TOGETHER:
            if diagram[index + 1] == C.STR_BODY_BEGIN:
                together = True
                together_paused = False
            continue


def parse_for_links(diagram: list, filename: str) -> None:
    detected_name = ""

    collecting = False
    first = False
    together = False
    together_paused = False

    for index, item in enumerate(diagram):

        if collecting:
            if first:
                first = False
                # skip name
                continue
            if item == C.STR_BODY_BEGIN:
                if together:
                    together_paused = together
                    together = False
                continue
            elif item == C.STR_BODY_END:
                if together:
                    together = False
                    continue

                collecting = False

                if together_paused:
                    together = together_paused
                    together_paused = False
            continue

        if item.casefold() in C.VALID_KEYWORDS:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collecting = True
                detected_name = diagram[index + 1]
                anchorsuffix = d.get_topic_anchor_suffix(filename, item, detected_name)
                dic_of_links[detected_name] = "sec-" + anchorsuffix
                first = True
            continue

        if item.casefold() == C.KEYWORD_SPECIAL_TOGETHER:
            if diagram[index + 1] == C.STR_BODY_BEGIN:
                together = True
                together_paused = False
            continue


def parse_for_relationships(diagram: list, filename: str) -> None:
    parent_candidate = ""
    user_candidate = ""
    connection_candidate = ""

    collecting = False
    first = False
    together = False
    together_paused = False

    for index, item in enumerate(diagram):

        if collecting:
            if first:
                first = False
                # skip name
                continue
            if item == C.STR_BODY_BEGIN:
                if together:
                    together_paused = together
                    together = False
                continue
            elif item == C.STR_BODY_END:
                if together:
                    together = False
                    continue

                collecting = False

                if together_paused:
                    together = together_paused
                    together_paused = False
            continue

        if item.casefold() in C.VALID_KEYWORDS:
            if diagram[index + 2] == C.STR_BODY_BEGIN:
                collecting = True
                first = True
            continue

        if item.casefold() == C.KEYWORD_SPECIAL_TOGETHER:
            if diagram[index + 1] == C.STR_BODY_BEGIN:
                together = True
                together_paused = False
            continue

        if item == C.UML_RELATION_EXTENSION_1 or item == C.UML_RELATION_EXTENSION_2:
            suspect = diagram[index + 1]
            f.logprint(f"{suspect} has parent {parent_candidate}")
            if suspect in dic_of_parents:
                dic_of_parents[suspect].append(parent_candidate)
            else:
                dic_of_parents[suspect] = [parent_candidate]

            if parent_candidate in dic_of_children:
                dic_of_children[parent_candidate].append(suspect)
            else:
                dic_of_children[parent_candidate] = [suspect]
        else:
            parent_candidate = item

        if item == C.UML_RELATION_COMPOSITION_1 or item == C.UML_RELATION_COMPOSITION_2 or \
           item == C.UML_RELATION_AGGREGATION_1 or item == C.UML_RELATION_AGGREGATION_2:
            suspect = diagram[index + 1]
            if suspect.startswith(C.STR_QUOTATION_MARK) and suspect.endswith(C.STR_QUOTATION_MARK):
                suspect = diagram[index + 2]
            f.logprint(f"{suspect} is used by {user_candidate}")
            if suspect in dic_of_users:
                dic_of_users[suspect].append(user_candidate)
            else:
                dic_of_users[suspect] = [user_candidate]
        else:
            user_candidate = item

        if item == C.UML_RELATION_CONNECTION_1 or item == C.UML_RELATION_CONNECTION_2 or \
           item == C.UML_RELATION_SMALL_CONNECTION_1 or item == C.UML_RELATION_SMALL_CONNECTION_2:
            suspect = diagram[index + 1]
            f.logprint(f"{suspect} has {item} connection to {connection_candidate}")
            if suspect in dic_of_connections:
                dic_of_connections[suspect].append(connection_candidate)
            else:
                dic_of_connections[suspect] = [connection_candidate]
        else:
            connection_candidate = item


def parse_items(items: list, name: str, filename: str) -> tuple:
    candidate = ""
    detected_normal = []
    detected_basics = []
    detected_states = []
    detected_methods = []
    detected_modifiers = []

    comment = False
    comment_type = C.ADOC_KEYWORD_NONE
    new_comment = False
    has_comment = False

    for i, item in enumerate(items):
        if item == C.STR_COMMENT_ADOC_REPLACEMENT:
            comment_type = C.ADOC_KEYWORD_DEFAULT
            comment = True
            new_comment = True
            has_comment = True
        elif item == C.STR_COMMENT_END_REPLACEMENT:
            if comment_type == C.ADOC_KEYWORD_PROPERTY or comment_type == C.ADOC_KEYWORD_DEFAULT:
                detected_normal.append("\n|")
            comment = False
            new_comment = False
            comment_type = C.ADOC_KEYWORD_NONE
        else:
            if comment:
                if item == C.ADOC_KEYWORD_STATE or \
                   item == C.ADOC_KEYWORD_PROPERTY or \
                   item == C.ADOC_KEYWORD_BASIC or \
                   item == C.ADOC_KEYWORD_METHOD or \
                   item == C.ADOC_KEYWORD_MODIFIER or \
                   item == C.ADOC_KEYWORD_TARGETPAGE or \
                   item == C.ADOC_KEYWORD_PACKAGENAME or \
                   item == C.ADOC_KEYWORD_TAG or \
                   item == C.ADOC_KEYWORD_IGNORE_FOR_OSC:
                    comment_type = item

                if comment_type == C.ADOC_KEYWORD_STATE:
                    if new_comment and detected_normal and len(items) > i and detected_normal[-1] in items[i + 1]:
                        f.log_error(f"state mismatch for '{detected_normal[-1]}' of {name}", filename)
                        detected_normal.pop() # there should be no normal elements for states
                    if new_comment:
                        new_comment = False
                    detected_states.append(item)
                elif comment_type == C.ADOC_KEYWORD_BASIC:
                    if new_comment:
                        new_comment = False
                    detected_basics.append(item)
                elif comment_type == C.ADOC_KEYWORD_METHOD:
                    if new_comment:
                        new_comment = False
                    else:
                        detected_methods.append((candidate, item))
                elif comment_type == C.ADOC_KEYWORD_MODIFIER:
                    if new_comment:
                        new_comment = False
                    else:
                        detected_modifiers.append((candidate, item))
                elif comment_type == C.ADOC_KEYWORD_TARGETPAGE:
                    if new_comment:
                        new_comment = False
                    else:
                        pass #already processed
                elif comment_type == C.ADOC_KEYWORD_PACKAGENAME:
                    if new_comment:
                        new_comment = False
                        try:
                            dic_of_packagenames[name] = items[i + 1]
                        except:
                            f.log_error("Invalid package name")
                    else:
                        pass
                elif comment_type == C.ADOC_KEYWORD_TAG:
                    if new_comment:
                        new_comment = False
                        try:
                            dic_of_tagnames[name] = items[i + 1]
                        except:
                            f.log_error("Invalid tag name")
                    else:
                        pass
                elif comment_type == C.ADOC_KEYWORD_IGNORE_FOR_OSC:
                    if new_comment:
                        new_comment = False
                        try:
                            dic_of_oscignore[name] = name
                        except:
                            f.log_error("Invalid name to ignore")
                else:
                    if new_comment:
                        detected_normal.append(candidate)
                        detected_normal.append(C.STR_TABLE_SEPARATOR)
                        new_comment = False
                    detected_normal.append(item) #also for property
            else:
                if not detected_normal:
                    detected_normal.append("\n|")
                candidate = item
                #detected_normal.append(f"{item}")

    if detected_normal:
        if has_comment:
            detected_normal = [e for e in detected_normal if e != C.STR_NEWLINE_REPLACEMENT]
            detected_normal.pop()
        else:
            if C.STR_NEWLINE_REPLACEMENT in detected_normal:
                detected_normal.pop(0)
                for i, e in enumerate(detected_normal):
                    if e == C.STR_NEWLINE_REPLACEMENT:
                        rep = f"{C.STR_NEWLINE}{C.STR_TABLE_SEPARATOR}"
                        if i != 0:
                            rep = f"{C.STR_TABLE_SEPARATOR}" + rep
                        detected_normal[i] = rep
            else:
                detected_normal.append("|")

    if detected_basics:
        detected_basics = [e for e in detected_basics if e != C.STR_NEWLINE_REPLACEMENT]

    detected_source = parse_source(items)

    return (detected_normal, detected_basics, detected_states, detected_source, detected_methods, detected_modifiers)


def parse_items_for_targetpages(items: list, name: str):

    comment = False
    comment_type = C.ADOC_KEYWORD_NONE
    new_comment = False

    for i, item in enumerate(items):
        if item == C.STR_COMMENT_ADOC_REPLACEMENT:
            comment_type = C.ADOC_KEYWORD_DEFAULT
            comment = True
            new_comment = True
        elif item == C.STR_COMMENT_END_REPLACEMENT:
            comment = False
            new_comment = False
            comment_type = C.ADOC_KEYWORD_NONE
        else:
            if comment:
                if item == C.ADOC_KEYWORD_TARGETPAGE or \
                   item == C.ADOC_KEYWORD_PACKAGENAME or \
                   item == C.ADOC_KEYWORD_TAG or \
                   item == C.ADOC_KEYWORD_IGNORE_FOR_OSC:
                    comment_type = item

                if comment_type == C.ADOC_KEYWORD_TARGETPAGE:
                    if new_comment:
                        new_comment = False
                    else:
                        dic_of_targetpages[name] = item
                else:
                    if new_comment:
                        new_comment = False

    dic_of_parentproperties[name] = parse_source(items)

    return


def parse_source(rawitems : list) -> list:
    detected = []

    items = [e for e in rawitems if e != C.STR_NEWLINE_REPLACEMENT]
    for index, item in enumerate(items):
        if item == C.ADOC_KEYWORD_STATE:
                try:
                    state_name = C.STR_STATE_PREFIX + items[index + 1].split(":")[0]
                    state_type = items[index + 2].split(";")[0]
                    state_mand = items[index + 2].split(";")[1]
                    state_expl = items[index + 2].split(";")[2]
                    # collect rest of the comment
                    ii = index + 3
                    while ii < len(items):
                        if items[ii] == C.STR_COMMENT_END_REPLACEMENT:
                            break
                        state_expl = state_expl + " " + items[ii]
                        ii+=1
                    detected.append((state_name, state_type, state_expl, state_mand))
                except:
                    f.log_error("Could not resolve states")
        elif item == C.ADOC_KEYWORD_PROPERTY:
                try:
                    prop_name = items[index - 2]
                    prop_type = items[index + 1].split(";")[0]
                    prop_mand = items[index + 1].split(";")[1]
                    prop_expl = items[index + 1].split(";")[2]
                    prop_dflt = ""
                    # collect rest of the comment
                    ii = index + 2
                    while ii < len(items):
                        if items[ii] == C.STR_COMMENT_END_REPLACEMENT:
                            break
                        opt_pos = items[ii].find(";")
                        if opt_pos > -1:
                            prop_expl = prop_expl + " " + items[ii][:opt_pos]
                            prop_dflt = items[ii][opt_pos + 1:]
                            ii+=1
                            break #end expl, start dflt
                        else:
                            prop_expl = prop_expl + " " + items[ii]
                        ii+=1
                    while ii < len(items):
                        if items[ii] == C.STR_COMMENT_END_REPLACEMENT:
                            break
                        prop_dflt = prop_dflt + " " + items[ii]
                        ii+=1

                    detected.append((prop_name, prop_type, prop_expl, prop_mand, prop_dflt.lstrip()))
                except:
                    f.log_error("Could not resolve properties")

    if not detected:
        detected.append((C.STR_EMPTY_DETAILS, C.STR_EMPTY_DETAILS, C.STR_EMPTY_DETAILS, C.STR_EMPTY_DETAILS))

    return detected


def register_stereotype(name : str, stereotype: str) -> None:
    stereotype = stereotype.lstrip("<<").rstrip(">>").capitalize()
    if name in dic_of_stereotypes and dic_of_stereotypes[name] != stereotype:
        f.log_error(f"Stereotype for \"{name}\" already exists. Was \"{dic_of_stereotypes[name]}\", now \"{stereotype}\"")
    dic_of_stereotypes[name] = stereotype


def get_dic_of_stereotypes() -> dict:
    return dic_of_stereotypes


def get_set_of_lists() -> set:
    return (list_of_classes, list_of_enums, list_of_entities, list_of_interfaces, list_of_abstracts)


def get_set_of_dictionaries() -> set:
    return (dic_of_children, dic_of_connections, dic_of_links, dic_of_parents, dic_of_users)


def get_dic_of_links() -> dict:
    return dic_of_links


def get_dic_of_targetpages() -> dict:
    return dic_of_targetpages


def get_dic_of_packagenames() -> dict:
    return dic_of_packagenames


def get_dic_of_tagnames() -> dict:
    return dic_of_tagnames


def get_dic_of_oscignore() -> dict:
    return dic_of_oscignore


def get_dic_of_parentproperties() -> dict:
    return dic_of_parentproperties
