# -*- coding: utf-8 -*-
import os
import sys
import subprocess

import yaml

from helpers import constants as C
from helpers import filehelper as f
from helpers import virtualfilehelper as v
from helpers import sanitizer as s
from helpers import uml_parser as p
from helpers import dumper as d

# Main
if __name__ == "__main__":
    config_ok = False
    use_virtual_files = False
    if len(sys.argv) == 2:
        if sys.argv[1] == "-h" or sys.argv[1] == "--help":
            # no logprint because we do not know where to log
            print(C.VERSION)
            print("This script generates adoc-fragments and osc2-source-fragments by analyzing comments in plantuml files.\n")
        else:
            config_ok = f.load_config(sys.argv[1])
    elif len(sys.argv) == 3:
        if sys.argv[1] == "-h" or sys.argv[1] == "--help" or sys.argv[2] == "-h" or sys.argv[2] == "--help":
            # no logprint because we do not know where to log
            print(C.VERSION)
            print("This script generates adoc-fragments and osc2-source-fragments by analyzing comments in plantuml files.\n")
        else:
            config_ok = f.load_config(sys.argv[1])
            virtual_data = v.load_files(sys.argv[2])
            if virtual_data:
                use_virtual_files = True

    else:
        config_ok = f.load_config("./uml2adoc_config.yml")

    if not config_ok:
        print("Please provide a config.yml file as an argument!")
        print("Call: python uml2adoc.py <config.yml>")
        sys.exit(-1)

    f.prepare_filestructure()
    f.logprint(C.VERSION)

    if use_virtual_files:
        f.set_dic_config(v.update_dic_config(f.get_dic_config()))

    try:
        if f.get_plantuml_jar():
            f.logprint("plantuml starting...")
            subprocess.run(["java", "-jar", f.get_plantuml_jar(), f.get_uml_inpath(), "-o", f.get_png_outpath(), "-tsvg", "-gui"])
            f.logprint("plantuml finished")
    except:
        # not an error because svg generating is optional
        f.logprint("no plantuml.jar defined or found")

    d.paste_bare(use_virtual_files)

    f.logprint("Parse 1...")
    if use_virtual_files:
        for v_file in v.get_virtual_files_with_path(f.get_uml_inpath())['fileList']:
            file = v_file['src']['basename']
            filename = v_file['src']['stem']
            file_extension = v_file['src']['extname']
            if ((file_extension == C.FILE_ENDING_PLANTUML) or (file_extension == C.FILE_ENDING_UML)):
                    lines = bytes(v_file['_contents']['data']).decode('utf-8').split("\n")
                    diagram = s.sanitize(lines, filename)

                    if not C.ADOC_KEYWORD_IGNORE in diagram:
                        p.parse_uml_diagram_for_links(diagram, filename, file)

    else:
        for root, dirs, files in os.walk(f.get_uml_inpath()):

            for file in files:
                if ((file.find(C.FILE_ENDING_PLANTUML) > -1) or (file.find(C.FILE_ENDING_UML) > -1)):
                    filename = os.path.splitext(file)[0]

                    with open(f"{root}/{file}", 'r', encoding="utf-8") as fobj:
                        lines = fobj.readlines()

                    diagram = s.sanitize(lines, filename)

                    if not C.ADOC_KEYWORD_IGNORE in diagram:
                        p.parse_uml_diagram_for_links(diagram, filename, file)

    f.logprint("Parse 2...")
    if use_virtual_files:
        for v_file in v.get_virtual_files_with_path(f.get_uml_inpath())['fileList']:
            file = v_file['src']['basename']
            filename = v_file['src']['stem']
            file_extension = v_file['src']['extname']
            if ((file_extension == C.FILE_ENDING_PLANTUML) or (file_extension == C.FILE_ENDING_UML)):
                    lines = bytes(v_file['_contents']['data']).decode('utf-8').split("\n")
                    diagram = s.sanitize(lines, filename)

                    if not C.ADOC_KEYWORD_IGNORE in diagram:
                        p.parse_uml_diagram(diagram, filename, file, True)
    else:
        for root, dirs, files in os.walk(f.get_uml_inpath()):

            for file in files:
                if ((file.find(C.FILE_ENDING_PLANTUML) > -1) or (file.find(C.FILE_ENDING_UML) > -1)):
                    f.logprint("\nparsing uml: " + file)
                    filename = os.path.splitext(file)[0]

                    with open(f"{root}/{file}", 'r', encoding="utf-8") as fobj:
                        lines = fobj.readlines()

                    diagram = s.sanitize(lines, filename)

                    # f.logprint("=" * 60)
                    # f.logprint(diagram)
                    # f.logprint("=" * 10)

                    if not C.ADOC_KEYWORD_IGNORE in diagram:
                        p.parse_uml_diagram(diagram, filename, file)

    d.dump_standard_osc(use_virtual_files)
    #d.dump_lists(use_virtual_files)
    if use_virtual_files:
        path = "/".join(sys.argv[2].split("/")[:-1])+"/"
        d.virtual_dump(path)
    f.log_summary()
